//
// Classe per manipolare stringhe.
//

#include <iostream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <string.h>
#include "String.H"


// Construttore: crea stringhe da un array di caratteri
string::string(const char *a)
{
  len = strlen(a);		// La lunghezza della stringa
  if (len == 0)
    p = 0;
  else {
    p = new char[1 + len];	// Alloco spazio sufficente per la stringa
    *p = 1;			// Metto ad 1 il contatore di riferimenti
    memcpy(p + 1, a, len);	// Copio la stringa nel nostro array, lasciando
  }				// nella prima posizione il contatore.
}


// Copy constructor
string::string(const string &t)
{
  len = t.len;			// Salvo la lunghezza della stringa

  if (t.p == NULL)		// Se la stringa e` vuota non ho altro da fare
    p = NULL;
  else if (*t.p != RC_MAX) {	// Se non ho raggiunto il numero massimo di
    (*t.p)++;			// riferimenti, aumento di uno il contatore
    p = t.p;			// e copio il puntatore alla stessa stringa
  }
  else {			// Devo creare effettivamente la nuova stringa
    p = new char[1 + len];	// Alloco lo spazio
    *p = 1;			// Metto ad 1 il contatore di riferimenti
    memcpy(p + 1, t.p + 1, len);// Copio la stringa nell'array
  }
}



// Distruttore
string::~string()
{
  if ((p != NULL) && (--*p == 0))	// Se la stringa non e` vuota e se non
    delete[] p;				// ho piu` riferimenti, libero la mem.
}



// Concatenazione di stringhe
string operator+ (const string &s, const string &t)
{
  if (s.len == 0)	// Se la prima stringa e` nulla ritorno la seconda
    return(t);					
  else if (t.len == 0)	// Se la seconda stringa e` nulla ritorno la prima
    return(s);
  else {		// Devo costruire una nuova stringa
    string u;

    u.len = s.len + t.len;		// La lunghezza e` data dalla somma
    u.p = new char[1 + s.len + t.len];	// Alloco lo spazio
    *u.p = 1;				// Ho un solo riferimento
    memcpy(u.p+1, s.p + 1, s.len);	// Copio la prima stringa nell'array
    memcpy(u.p+1 + s.len, t.p + 1, t.len);	// Copio la seconda stringa

    return(u);				// Restituisco la nuova stringa
  }
}


// Test di eguaglianza
int operator== (const string &s, const string &t)
{
  if (s.len != t.len)		// Se le due lunghezze sono diverse, le
    return(0);			// stringhe sono diverse
  else 
    return(memcmp(s.p + 1, t.p + 1, s.len) == 0);	// Confronto byte a byte 
}


// Test di precedenza in ordine lessicografico
int operator<= (const string &s, const string &t)
{
  int m = (s.len <= t.len) ? s.len : t.len;	// Determino la maggiore delle
						// lunghezze
  int c = memcmp(s.p + 1, t.p + 1, m);	// Confronto i primi m caratteri

  if (c != 0)		// Se non sono uguali i primi m, ritorno il segno di c
    return(c < 0);	// ovvero 1 solo se s e` minore di t
  else
    return(s.len <= t.len);	// Ritorna 1 solo s e` piu` corta di t
}


// Date 2 stringhe, scorre la prima stringa per la prima occorrenza della
// seconda stringa
int pos(const string &s, const string &t, int i)
{
  if ((i < 1) || (i > s.len))	// Se cerco una sottostringa fuori dal range
    return(0);			// sicuramente non la trovo

  int imax = s.len - t.len + 1;	// Al piu` la seconda stringa potra` cominciare
				// a t.len caratteri dalla fine della prima
  for( ; i <= imax; i++)	// Per ogni carattere compreso fra i e imax
    if(memcmp(s.p + i, t.p + 1, t.len) == 0)	// Verifico se la seconda
      return(i);				// stringa comapre nella prima

  return(0);			// Se arrivo qui non l'ho trovata.
}


// Assegnamento
string& string::operator= (const string &t)
{
  len = t.len;			// Assegno la lunghezza della stringa

  if (this != &t) {		// Verifico se assegno a se stesso
    if ((p != NULL) && (--*p == 0))	// Se la stringa a cui assegno non e`
      delete[] p;			// la stringa vuota, ma non ci punta
					// nulla, libero la memoria. Comunque
					// decremento il contatore di riferim.
    if (t.p == NULL)		// Se e` vuota la stringa originale, assegno la
      p = NULL;			// stringa vuota a p.
    else if (*t.p != RC_MAX) {
      (*t.p)++;			// Se non ho raggiunto il numero massimo di
				// riferimenti, incremento il contatore.
      p = t.p;			// Poi copio il puntatore.
    }
    else {			// Ho raggiunto il massimo numero di rifer.
      p = new char[1 + len];		// Alloco la nuova memoria
      *p = 1;				// Metto ad 1 il contatore di riferim.
      memcpy(p + 1, t.p + 1, len);	// Copio la stringa
    }
  }

  return(*this);		// Restituisco la stringa come risultato
}


// Concatenazione ed assegnamento
string& string::operator+= (const string &t)
{
  if (t.len != 0) {		// Se la stringa da concatenare non e` vuota
    int new_len = len + t.len;	// Concateno le stringhe creandone una nuova 
    char *new_p = new char[1 + new_len];
    *new_p = 1;
    memcpy(new_p + 1, p + 1, len);
    memcpy(new_p + 1 + len, t.p + 1, t.len);
    len = new_len;

    if ((p != NULL) && (--*p == 0))	// Libero eventualmente la memoria
      delete[] p;

    p = new_p;			// Copio il puntatore
    }

  return(*this);		// Restituisco la stringa come risultato
}


// Subscripting: voglio accedere alla posizione i-esima della stringa.
// Se ho piu` riferimenti alla stessa stringa dovro` creare un nuovo oggetto e 
// copiarci dentro la vecchia stringa. In questo modo se modifico il carattere
// nella posizione i-esima lo faccio solo per questa stringa e non per tutte
char& string::operator[] (int i)
{
  if ((i < 1) || (i > len)) {			// Controllo se l'indice e`
    cerr << "Suscript out of range" << endl;	// nell'intervallo corretto
    exit(1);
  }

  if (*p > 1) {				// Se ho piu` di un riferimento devo
    char *q = new char[1 + len];	// Creare una nuova stringa
    *q = 1;				// Mettere ad 1 il contatore di rif.
    memcpy(q + 1, p + 1, len);		// Copiare la stringa
    --*p;				// Decrementare il contatore di rif. 
    p = q;				// Asseganre a p la nuova stringa
  }

  return(p[i]);				// Ritorno il riferimento all'elemento
}



// Estrae e ritorna la sottostringa specificata partendo dalla posizione i e
// lunga n
string string::operator() (int i, int n) const
{
  // Controllo se l'inizio della sottostringa e` nell'intervallo corretto
  if ((i < 1) || (i > len)) {
    cerr << "Substring position out of range" << endl;
    exit(1);						
  }

  // Verifico che la sottostringa abbia una lunghezza positiva
  if (n < 0) {
    cerr << "Negative substring length" << endl;
    exit(1);
  }

  string sub;			// Creo la sottostringa nulla

  if (n == 0) 
    return(sub);		// Restituisce la stringa nulla

  if (n > len - i + 1)		// Se ho meno di n caratteri a partire dalla
    n = len - i + 1;		// posizione i, copio solo fino alla fine della
				// stringa originale
  sub.len = n;			// Assegno la lunghezza della sottostringa
  sub.p = new char[1 + n];	// Alloco lo spazio
  *sub.p = 1;			// Metto ad 1 il contatore di rif.
  memcpy(sub.p + 1, p + i, n);	// Copio la sottostringa

  return(sub);			// Ritorno la nuova sottostringa.
}



// Operatore di estrazione da stream
istream& operator>> (istream &ci, string &t)
{
  char buffer[256];

  cin >> setw(256) >> buffer;	// L'input e` troncato al 256-esimo carattere
  t = string(buffer);		// La stringa cosi` letta e` assegnata a t

  return(ci);
}


// Operatore di inserimento in uno stream
ostream& operator<< (ostream &co, const string &t)
{
  for(int i = 1; i <= t.len; i++)	// Inserisce un carattere per volta
    co << t.p[i];			// nell'output stream

  return(co);
}

